HealInc = LibStub("AceAddon-3.0"):NewAddon("AllSeeingEye", "AceConsole-3.0","AceEvent-3.0","LibBars-1.0")
										 
local HealComm 	= LibStub("LibHealComm-3.0")
local ResComm 	= LibStub("LibResComm-1.0")
local QuickHealth = LibStub("LibQuickHealth-2.0")
local Bars = LibStub("LibBars-1.0")
local Media = LibStub("LibSharedMedia-3.0")
local DataBroker = LibStub:GetLibrary("LibDataBroker-1.1",true)

local playerName = UnitName('player');
local currentTarget;

local UnitHealth = QuickHealth and QuickHealth.UnitHealth or UnitHealth

local hexcolors = { 
	PRIEST = "FFFFFF",
	SHAMAN = "2459FF",
	PALADIN = "F58CBA",
	DRUID = "FF7D0A",
	DEATHKNIGHT = "C41F3B",
	HUNTER = "ABD473",
	MAGE = "69CCF0",
	ROGUE = "FFF569",
	WARLOCK = "9482C9",
	WARRIOR = "C79C6E",
}

local colors = {
	green = {0,1,0},
	orange = {1,0.51,0},
	red = {1,0,0}
}
-- -------------------------------------------------------- --
-- 	    Slash commands, defaults and general settings 		--
-- -------------------------------------------------------- --	
function HealInc:OnInitialize()	
	
	local options = { 
        type = 'group',
            args = {
					spacer1 = {
						type = "header",
						order = 1,
						name = "Bar Style",
					},
                    anchor = {
    				    order = 2,
    	       			type = 'toggle',
	       	      		name = "Lock",
		      	     	desc = "Show / Hide anchors",
			     	    get = function() return not self.heal_bars:IsAnchorVisible() end,
				        set = function(info,val) 
							if val then 
								self.heal_bars:HideAnchor(); self.res_bars:HideAnchor(); 
								self.db.profile.locked = true 
							else 
								self.heal_bars:ShowAnchor(); self.res_bars:ShowAnchor(); 
								self.db.profile.locked = false 
							end 
						end,
						width = "full"
                    },
				    scale = {
						order = 3,
						type = 'range',
						name = "Bar Scale",
						desc = "Changes the scale of bars.",
						get = function() return self.db.profile.scale  end,
						set = function(info,val) 
							self.heal_bars:SetScale(val)
							self.res_bars:SetScale(val)
							self.db.profile.scale = val 
						end,
						min = 0.5,
						max = 2,	
						step = 0.05,
						
			         },
					 texture = {
						type = 'select',
						name = 'Texture',
						desc = 'Set the bar Texture',
						values = function() 
							local list = {}
							for k, v in pairs(Media:List('statusbar')) do
								list[v] = v
							end 
							return list
						end,
						get = function() return self.db.profile.texture  end,
						set = function(info,val) 
							self.db.profile.texture = val 
							self.heal_bars:SetTexture(Media:Fetch("statusbar", val))
							self.res_bars:SetTexture(Media:Fetch("statusbar", val))
							end,
						order = 4,
					},
					grow = {
						type = 'toggle',
						name = 'Grow Upwards',
						desc = 'Toggling this will make the bars grow upwards',
						get = function() return self.db.profile.reverse_growth  end,
						set = function(info,val) 
							self.db.profile.reverse_growth = val 
							self.heal_bars:ReverseGrowth(val)
							self.res_bars:ReverseGrowth(val)
							end,
						order = 5,
						width = "full"
					},
					spacer3 = {
						type = "header",
						order = 20,
						name = "LibHealComm-3.0 versions",
					},
                    guild = {
            			name = "View Guild",
			            type = "execute",
            			order = 21,
			            desc = "Print the versions of LibHealComm-3.0 used by guild members",
            			func = function() 
                            local versions = LibStub("LibHealComm-3.0"):GetGuildVersions();
                            for player, version in pairs(versions) do                    
                                DEFAULT_CHAT_FRAME:AddMessage(player .. ": " .. tostring(version));
                            end
                        end,
		            },
		            group = {
                        name = "View Group",
			            type = "execute",
            			order = 22,
			            desc = "Print the versions of LibHealComm-3.0 used by group members",
            			func = function() 
                            local versions = LibStub("LibHealComm-3.0"):GetRaidOrPartyVersions();
                            for player, version in pairs(versions) do                    
                                DEFAULT_CHAT_FRAME:AddMessage(player .. ": " .. tostring(version));
                            end
                        end
                    }, 
					spacer2 = {
						type = "header",
						order = 10,
						name = "Resurrection monitor",
					},
					enable = {
				        order = 11,
				        type  = 'toggle',
            			name  = "Enable Monitor",
				        desc  = "Activate Resurection Monitor",
            			get   = function() return self.db.profile.resmon end,
				        set   = function(info,val)	self.db.profile.resmon = val end
					}
			}
	}

    
	local defaults 	= {
		profile = {
			scale        		= 1,
			resmon              = true,
			locked 				= false,
			hbX 				= 470,
            hbY 				= 275,
			rbX 				= 470,
            rbY 				= 375,
			texture 			= "Blizzard",
			reverse_growth		= false
		}
    }
    
	LibStub("AceConfig-3.0"):RegisterOptionsTable("HealInc", options)
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("HealInc", "HealInc")
	self:RegisterChatCommand("healinc", function () InterfaceOptionsFrame_OpenToCategory(self.optionsFrame) end)
	self.message = "HealInc"

	self.db = LibStub("AceDB-3.0"):New("HealIncDB", defaults)
	
	if DataBroker then
		local launcher = DataBroker:NewDataObject("HealInc", {
		type = "launcher",
		icon = "Interface\\Icons\\Spell_Nature_HealingTouch",
		OnClick = function(clickedframe, button)
			if button == "LeftButton" then
				self.heal_bars:ToggleAnchor()
				self.res_bars:ToggleAnchor()
			elseif button == "RightButton" then
				InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
			end
		end,
		OnTooltipShow = function(self)
			GameTooltip:AddLine("HealInc "..GetAddOnMetadata("HealInc", "version"), HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
			GameTooltip:AddLine("Left click to lock/unlock the timer bars. Right click for configuration.", NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
			GameTooltip:Show()
		end,
		})		

	end
	
    self.heals = {}
    self.res = {}
    self.ressed = {}
    
	local db = self.db.profile
	
	self.heal_bars = self:NewBarGroup("HealInc", Bars.RIGHT_TO_LEFT)
	self.heal_bars:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", db.hbX, db.hbY)
    self.heal_bars:SetScale(db.scale)
	self.heal_bars:ReverseGrowth(db.reverse_growth)
	if db.locked then self.heal_bars:HideAnchor() else self.heal_bars:ShowAnchor() end
	
	self.res_bars = self:NewBarGroup("Res-monitor", Bars.RIGHT_TO_LEFT, 300)
	self.res_bars:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", db.rbX, db.rbY)
    self.res_bars:SetScale(db.scale)
	self.res_bars:ReverseGrowth(db.reverse_growth)
	if db.locked then self.res_bars:HideAnchor() else self.res_bars:ShowAnchor() end
	
	self.heal_bars:SetSortFunction( function(a,b) return a.value < b.value end )
	
	--self.heals["Foobar"] = {}
	--self.heals["Foobar"]["Testhealer"] = {}
	
	--self.heals["Foobar"]["Testhealer"].endTime = 5183.805
	--self.heals["Foobar"]["Testhealer"].size = 20
	--self.heals["Foobar"]["Testhealer"].startTime = 3183.805
end

function HealInc:OnEnable()
	self:RegisterEvent("PLAYER_TARGET_CHANGED", "TargetChanged");

	--self:RegisterEvent("UNIT_HEALTH","TargetHealthChanged");
	QuickHealth.RegisterCallback(self, "UnitHealthUpdated",  "TargetHealthChanged");
	
	HealComm.RegisterCallback(self, "HealComm_DirectHealStart");
	HealComm.RegisterCallback(self, "HealComm_DirectHealDelayed");
	HealComm.RegisterCallback(self, "HealComm_DirectHealStop");
	
	ResComm.RegisterCallback(self, "ResComm_ResStart");
	ResComm.RegisterCallback(self, "ResComm_ResEnd");
	ResComm.RegisterCallback(self, "ResComm_Ressed");
	
	self.heal_bars.RegisterCallback(self,"AnchorMoved","HealAnchorMoved")
	self.res_bars.RegisterCallback(self, "AnchorMoved","ResAnchorMoved")
	
	-- Can only set the texture after SharedMedia is loaded
	self.heal_bars:SetTexture(Media:Fetch("statusbar", self.db.profile.texture))
	self.res_bars:SetTexture(Media:Fetch("statusbar", self.db.profile.texture))
end

function HealInc:OnDisable()
	self:UnregisterBucketEvent("UNIT_HEALTH");
    self:UnregisterEvent("PLAYER_TARGET_CHANGED");

    HealComm.UnregisterCallback(self, "HealComm_DirectHealStart");
    HealComm.UnregisterCallback(self, "HealComm_DirectHealDelayed");
    HealComm.UnregisterCallback(self, "HealComm_DirectHealStop");
    
    ResComm.UnregisterCallback(self, "ResComm_ResStart");
	ResComm.UnregisterCallback(self, "ResComm_ResStop");
    ResComm.UnregisterCallback(self, "ResComm_Ressed");
end

----------------------------------------
-- EVENTS
----------------------------------------

function HealInc:HealAnchorMoved(_, _, x, y)
    self.db.profile.hbX, self.db.profile.hbY = x, y

end

function HealInc:ResAnchorMoved(_, _, x, y)
    self.db.profile.rbX, self.db.profile.rbY = x, y

end

function HealInc:HealComm_DirectHealStart(event, healerName, healSize, endTime, ...)
    for i=1,select('#', ...) do
           
        local targetName = select(i, ...);

        if not self.heals[targetName] then self.heals[targetName] = {} end
        
        local realSize = floor( HealComm:UnitHealModifierGet(targetName) * healSize);

        _, _, _, icon, _ = UnitCastingInfo(healerName);
               
        -- store the heal in internal heal tracking
        self.heals[targetName][healerName] = { endTime = endTime,
                                               size = realSize,
                                               startTime = GetTime(),
                                               icon = icon 
                                            }

		self:StartBar(healerName, targetName);
        
    end
	
	self:UpdateOverhealColors()
end

function HealInc:HealComm_DirectHealStop(event, healerName, healSize,  succeeded, ...)
    for i=1,select('#', ...) do
        local targetName = select(i, ...);
        
        -- Only change the text of a bar if not succeeded, it should run out naturally.
        if not succeeded then self:StopBar(healerName,targetName) end;
        
		-- remove data from heal table
        self.heals[targetName][healerName] = nil;
        
    end
	self:UpdateOverhealColors()
end

function HealInc:HealComm_DirectHealDelayed(event, healerName, healSize, endTime, ...)
    for i=1,select('#', ...) do
        local targetName = select(i, ...);
             
        self.heals[targetName][healerName].endTime = endTime;
        
        self:DelayBar(healerName,targetName);
    end
	self:UpdateOverhealColors()
end

function HealInc:TargetChanged()
	currentTarget = UnitName("target")
	
    self:UpdateBars();
    
    self:UpdateOverhealColors()
end

function HealInc:TargetHealthChanged(event,unit)
    if unit ~= "target" then return end

    self:UpdateOverhealColors()
end

------------------------------------------
-- BARS
------------------------------------------

function HealInc:StartBar(healerName, targetName)
	if targetName ~= currentTarget then return end
	
    local info = self.heals[targetName][healerName]
    local id = "HealInc-"..healerName;
    local bartext = "";
    
    local totalTime = info.endTime - info.startTime;
    local time = info.endTime - GetTime();

    bartext = self:ClassTextColor(healerName,select(2, UnitClass(healerName))).." - "..info.size

	local bar = self.heal_bars:NewTimerBar(id, bartext, time, totalTime, info.icon, 0)
	r,g,b = unpack(colors.green)
	bar:SetBackgroundColor(r,g,b,1)
    
	bar:SetColorAt(0,0,0,0,1.0)

    -- register bar
    self.heals[targetName][healerName].bar = bar;
end

function HealInc:StopBar(healerName, targetName)
	if targetName ~= currentTarget then return end
	
    local info = self.heals[targetName][healerName]
    
	if not info then
		-- apparently the bar was already stopped. This should not happen., but happen on prayer of healing
		return
	end

	info.bar:Fade(0.5)
            
end

function HealInc:DelayBar(healerName,targetName)
	if targetName ~= currentTarget then return end
	
    local info = self.heals[targetName][healerName];
       
    local totalTime = info.endTime - info.startTime;
    local time = info.endTime - GetTime();
    
    info.bar:SetTimer(time, totalTime)
end
    
-- Called when target is changed
function HealInc:UpdateBars()

	if self.heal_bars:HasAnyBar() then
		for name,bar in self.heal_bars:IterateBars() do 
			self.heal_bars:RemoveBar(bar)
		end
	end
	    
    -- If no target selected, or no incomming heals on the new target then end function.
    if not currentTarget or not self.heals[currentTarget] then return end
        
    for healerName,healData in pairs(self.heals[currentTarget]) do
        self:StartBar(healerName,currentTarget);        
    end
end

-----------------------------------------------
-- OVERHEAL COLORS
-----------------------------------------------
function HealInc:UpdateOverhealColors()
	
    local currentHeals = {}
    
    if not currentTarget or not self.heals[currentTarget] then return end
    
    local currentHealth = UnitHealth("target");
    local maxHealth = UnitHealthMax("target");
    local newHealth = currentHealth;
    
    for healerName,info in pairs(self.heals[currentTarget]) do
        info.healerName = healerName;
        table.insert(currentHeals,info);
    end
    
    table.sort(currentHeals,function(a,b) return a.endTime < b.endTime end);

    for index,info in pairs(currentHeals) do
        
        -- Check if the heal is going to overheal.
        if (newHealth + (info.size * 0.75)) < maxHealth then
			r,g,b = unpack(colors.green)
            info.bar:SetBackgroundColor(r,g,b,1)
        elseif (newHealth + (info.size * 0.1)) < maxHealth then
            r,g,b = unpack(colors.orange)
            info.bar:SetBackgroundColor(r,g,b,1)  
        else
            r,g,b = unpack(colors.red)
            info.bar:SetBackgroundColor(r,g,b,1)
        end
            
        newHealth = newHealth + info.size;
    end
end

----------------------------------------------
-- FUNCTIONS
----------------------------------------------
function HealInc:ClassTextColor(text, class)
	if class and hexcolors[class] then
		return format("|cff%s%s|r", hexcolors[class], text)
	else
		return text
	end
end

